/****************************************************************************
*                   source\windows\config.h
*
*  This file contains the Windows specific defines.
*
*  from Persistence of Vision Raytracer
*  Copyright 1996-2002 Persistence of Vision Team
*  This windows version of POV-Ray is Copyright 1996-2002 Christopher J. Cason.
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided so that users may experiment
*  with enhancements to POV-Ray and to port the software to platforms other
*  than those supported by the POV-Ray Team.  There are strict rules under
*  which you are permitted to use this file.  The rules are in the file
*  named POVLEGAL.DOC which should be distributed with this file. If
*  POVLEGAL.DOC is not available it may be found online at the following URL:
*
*    http://www.povray.org/povlegal.html.
*
* This program is based on the popular DKB raytracer version 2.12.
* DKBTrace was originally written by David K. Buck.
* DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
*
* $File: //depot/povray/3.5/windows/config.h $
* $Revision: #33 $
* $Change: 1814 $
* $DateTime: 2002/07/27 10:19:15 $
* $Author: chrisc $
* $Log$
*
*****************************************************************************/

#ifndef CONFIG_H
#define CONFIG_H

// some params are of type COLOUR, which is not yet defined (it's in frame.h).
// this allows us to substitute _COLOUR for it within this file.
typedef float _COLOUR [5] ;

#define FILE_NAME_LENGTH          _MAX_PATH
#define POV_NAME_MAX              _MAX_FNAME

#define IFF_SWITCH_CAST (long)

#define POVRAY_PLATFORM_NAME "win32"

// NOTE NOTE NOTE - do NOT define PRECISION_TIMER_AVAILABLE unless you know
// what you are doing. and have read the comments in ZTIMER/READ.ME. You
// will need a working ZTimer library if this is switched on. The only thing
// you lose from not having it is the ability to do scene profiling.

#ifdef __WATCOMC__                            /* Watcom C/C++ C32 */
  #include <float.h>
  #define COMPILER_VER                        ".watcom"
  #include "ztimer.h"
  #define PRECISION_TIMER_AVAILABLE           1
  #define PRECISION_TIMER_INIT                ZTimerInit () ;
  #define PRECISION_TIMER_START               LZTimerOn() ;
  #define PRECISION_TIMER_STOP                LZTimerOff() ;
  #define PRECISION_TIMER_COUNT               LZTimerCount() ;
  #define QSORT_FUNCT_PARM                    const void _FAR *
  #define QSORT_FUNCT_RET                     int _Cdecl
  #define QSORT(a,b,c,d)                      qsort(a, b, c, d)
  #define FIX_WATCOM_BUG                      if (l<0) Fix_Watcom_Bug(temp4);
  #define HEAPSHRINK                          _heapshrink () ;
#endif

#ifdef __BORLANDC__                           /* Borland C/C++ */
  #define COMPILER_VER                        ".borland"
  #define IFF_SWITCH_CAST                     (long)
  #define QSORT_FUNCT_PARM                    const void *
  #define QSORT_FUNCT_RET                     int CDECL
  #include "ztimer.h"
  #define PRECISION_TIMER_AVAILABLE           1
  #define PRECISION_TIMER_INIT                ZTimerInit () ;
  #define PRECISION_TIMER_START               LZTimerOn() ;
  #define PRECISION_TIMER_STOP                LZTimerOff() ;
  #define PRECISION_TIMER_COUNT               LZTimerCount() ;
  #define MSDOS
  #define QSORT(a,b,c,d)                      qsort((void *) a, (size_t) b, (size_t) c, (int (_USERENTRY *) (const void *, const void *)) d)
#endif

#ifdef _MSC_VER                               /* Microsoft Visual C++ */ /* [LSK] */

#pragma warning(disable : 4018)	/* signed/unsigned mismatch */
  #pragma warning(disable : 4305)	/* truncation from 'type1' to 'type2' (mostly double to float) */
  #pragma warning(disable : 4244)	/* possible loss of data (converting ints to shorts) */

#ifdef __INTEL_COMPILER
  #define COMPILER_VER                        ".icl"
#ifdef INTELSSE2
  #define USE_INTEL_SSE2
  #define USE_FASTER_NOISE                    1
#endif
#else
  #define COMPILER_VER                        ".msvc"
#endif

  #define IFF_SWITCH_CAST                     (long)
  #define QSORT_FUNCT_PARM                    const void *
  #define QSORT_FUNCT_RET                     int CDECL
  #include "ztimer.h"
  #include "direct.h"
  #define PRECISION_TIMER_AVAILABLE           1
  #define PRECISION_TIMER_INIT
  #define PRECISION_TIMER_START               LZTimerOn() ;
  #define PRECISION_TIMER_STOP                LZTimerOff() ;
  #define PRECISION_TIMER_COUNT               LZTimerCount() ;
  #define MSDOS
  #define QSORT(a,b,c,d)                      qsort((void *) a, (size_t) b, (size_t) c, (int (__cdecl*) (const void *, const void *)) d)
  #define getcwd _getcwd

  // NB _heapmin () only works on NT.
  #define HEAPSHRINK                          _heapmin () ;

  #if (_MSC_VER<=1100)				/* this section for VC5 */
    #define RBBS_GRIPPERALWAYS                 0x00000080
    #define RBS_DBLCLKTOGGLE                   0x00008000
	#define TBSTYLE_TRANSPARENT				   0x00008000
	/* These two are not defined in VC5 */
	typedef long INT_PTR, *PINT_PTR;
	typedef unsigned long UINT_PTR, *PUINT_PTR;
  #endif

  #define NEED_INVHYP

  template<class _Ty> inline const _Ty& max(const _Ty& _X, const _Ty& _Y) {return (_X < _Y ? _Y : _X); }
  template<class _Ty> inline const _Ty& min(const _Ty& _X, const _Ty& _Y) {return (_Y < _X ? _Y : _X); }
#endif

#ifndef HEAPSHRINK
#define HEAPSHRINK
#endif

#ifndef NAN
#define NAN (10E100)
#endif

void Fix_Watcom_Bug(char *s);

// this is for other compilers that don't use ztimer.h
#ifndef __GENDEFS
#define __GENDEFS
typedef unsigned char   uchar;
typedef unsigned short  ushort;
typedef unsigned int    uint;
typedef unsigned long   ulong;
#endif

#ifndef COMPILER_VER
#define COMPILER_VER ".win_unsupported"
#endif

#include <stdarg.h>
#include <stdlib.h>

/* Un comment at most one of these options */
// #define MEM_TAG     // Enables memory tag debugging
// #define MEM_TRACE   // Enables garbage collection and memory tracing
#define MEM_RECLAIM    // Enables memory reclaimation

/* For now, don't allow POV to track C++ new and deletes */
#define	POV_USE_NATIVE_CPP_MEM
#define POV_CPP_MEM_HAS_NOTHROW_SUPPORT 0

#define RED_RAW_FILE_EXTENSION    ".r8"  /* PICLAB's "RAW" file format */
#define GREEN_RAW_FILE_EXTENSION  ".g8"  /* PICLAB's "RAW" file format */
#define BLUE_RAW_FILE_EXTENSION   ".b8"  /* PICLAB's "RAW" file format */
#define PLATFORM_CONSTANT         1.622
#define FILENAME_SEPARATOR        '\\'
#define DEFAULT_OUTPUT_FORMAT     's'
#define POV_IS1                   ".bmp"
#define POV_IS2                   ".BMP"
#define POV_NUM_ECHO_LINES        2
#define POV_ERROR_MARKER_TEXT     " <----ERROR\n"

// we do our own I/O Restrictions error handling
#define NO_GENERIC_IO_ERROR

int parse_ini_file(char *filename);

void WIN_Banner(char *s);
void WIN_Warning(char *s);
void WIN_Render_Info(char *s);
void WIN_Status_Info(char *s);
void WIN_Debug_Info(char *s);
void WIN_Fatal(char *s);
void WIN_Statistics(char *s);
void WIN_Startup(void);
void WIN_Finish(int n);
void WIN_Cooperate(int n);
int  WIN_Display_Init(int width, int height);
void WIN_Display_Plot(int x, int y, int r, int g, int b, int a);
void WIN_Display_Plot_Rect(int x1, int x2, int y1, int y2, int r, int g, int b, int a);
void WIN_Pre_Pixel (int x, int y, _COLOUR colour);
void WIN_Post_Pixel (int x, int y, _COLOUR colour);
void WIN_Display_Finished(void);
void WIN_Display_Close(void);
void WIN_LoadRerun(void);
void WIN_SaveRerun(void);
void WIN_ParseINI(void);
void WIN_PrintOtherCredits (void) ;
void WIN_Write_Line (_COLOUR *line, int y) ;
void WIN_Assign_Pixel (int x, int y, _COLOUR colour) ;
int WIN_System(char *s);
int WIN_Povray (int argc, char **argv) ;
void WIN_WhereError (char *FileName, int Line, char *SrcLine, int Index, char *Token) ;
int WIN_Allow_File_Write (const char *Filename, const unsigned int FileType) ;
int WIN_Allow_File_Read (const char *Filename, const unsigned int FileType) ;
int WIN_Shellout (int) ;

#define STARTUP_POVRAY            WIN_Startup();
#define FINISH_POVRAY             /* do nothing! */  /* called by povray_terminate */
#define EXIT_POVRAY(n)            WIN_Finish(n);     /* called by povray_exit */
#define COOPERATE_0               WIN_Cooperate(0);
#define COOPERATE_1               WIN_Cooperate(1);
#define TEST_ABORT                WIN_Cooperate(0);

#define POV_BANNER                WIN_Banner
#define POV_WARNING               WIN_Warning
#define POV_RENDER_INFO           WIN_Render_Info
#define POV_STATUS_INFO           WIN_Status_Info
#define POV_DEBUG_INFO            WIN_Debug_Info
#define POV_FATAL                 WIN_Fatal
#define POV_STATISTICS            WIN_Statistics

#define POV_DISPLAY_INIT(prc,w,h) WIN_Display_Init((w),(h));
#define POV_DISPLAY_CLOSE         WIN_Display_Close();
#define POV_DISPLAY_FINISHED      WIN_Display_Finished();
#define POV_DISPLAY_PLOT(prc,x,y,r,g,b,a) WIN_Display_Plot((x),(y),(r),(g),(b),(a));
#define POV_DISPLAY_PLOT_RECT(prc,x1,x2,y1,y2,r,g,b,a) WIN_Display_Plot_Rect((x1),(x2),(y1),(y2),(r),(g),(b),(a));
#define POV_PRE_PIXEL(x,y,c)      WIN_Pre_Pixel (x, y, c) ;
#define POV_POST_PIXEL(x,y,c)     WIN_Post_Pixel (x, y, c) ;
#define POV_SYSTEM(s)             WIN_System (s) ;
#define POV_WRITE_LINE(line,y)    WIN_Write_Line (line, y) ;
#define POV_ASSIGN_PIXEL(x,y,colour) WIN_Assign_Pixel (x, y, colour) ;
#define POV_WHERE_ERROR(fn,ln,ep,ei,ts) WIN_WhereError (fn, ln, ep, ei, ts) ;
#define POV_SHELLOUT(type)        WIN_Shellout (type)

#define POV_ALLOW_FILE_WRITE(f,t) WIN_Allow_File_Write (f, t)
#define POV_ALLOW_FILE_READ(f,t)  WIN_Allow_File_Read (f, t)

#define POV_BYTE_ORDER            'HHHH'

#define READ_ENV_VAR              if ((Option_String_Ptr = getenv("POVINI")) != NULL) parse_ini_file(Option_String_Ptr);
#define ALT_WRITE_INI_FILE        WIN_SaveRerun() ;
#define PROCESS_POVRAY_INI        WIN_ParseINI() ;

#define ALTMAIN
#define RENAME_FILE(orig,new)     rename(orig,new)
#define NEW_LINE_STRING           "\r\n"
#define SYS_IMAGE_CLASS           BMP_Image
#define SYS_IMAGE_HEADER          "pvbmp.h"
#define READ_SYS_IMAGE(i,n)       Read_BMP_Image((i),(n))
#define GET_SYS_FILE_HANDLE       Get_BMP_File_Handle
#define SYS_DEF_EXT               ".bmp"

#define POVMSBool bool

#define PRINT_OTHER_CREDITS       WIN_PrintOtherCredits () ;

/*
 VC6 math.h defines HUGE_VAL as +inf.  The bounding box routines 
 seem to have problems with +/- inf that lead to page faults.
 Rather than fix the bbox routines, which shouldn't have to deal
 with infinities anyway, use the frame.h value for HUGE_VAL 
 instead of the one in math.h   This depends on config.h being
 included after math.h, but that is the case in the current frame.h
 14 Jan 2001 RLP
*/
#undef HUGE_VAL

#define MEM_STATS                 1

// this is necessary to fix a problem with VC++ seeking to the
// wrong location during #while loops. ("r" isn't enough).
#define READ_TXTFILE_STRING "rb"

typedef struct Image_Struct IMAGE;
struct Image_Struct;

extern void Read_BMP_Image(IMAGE *Image, char *name);

#endif
